import matplotlib.pyplot as plt
import numpy as np

from ttim.aquifer import AquiferData
from ttim.aquifer_parameters import param_3d, param_maq
from ttim.linesink1d import FluxDiffLineSink1D, HeadDiffLineSink1D
from ttim.stripareasink import AreaSinkXsection, HstarXsection


class Xsection(AquiferData):
    """Base class for a cross-section inhomogeneity.

    Parameters
    ----------
    model : Model
        Model to add the cross-section to, usually an instance of ModelXsection.
    x1 : float
        x-coordinate of the left boundary of the cross-section.
    x2 : float
        x-coordinate of the right boundary of the cross-section.
    kaq : array
        Hydraulic conductivities of the aquifers.
    z : array
        Elevations of the tops and bottoms of the layers.
    Haq : array
        Thicknesses of the aquifers.
    Hll : array
        Thicknesses of the leaky layers.
    c : array
        Resistance of the leaky layers.
    Saq : array
        Specific storage of the aquifers.
    Sll : array
        Specific storage of the leaky layers.
    poraq : array
        Porosities of the aquifers.
    porll : array
        Porosities of the leaky layers.
    ltype : array
        Type of each layer. 'a' for aquifer, 'l' for leaky layer.
    topboundary : str
        Type of top boundary. Can be 'conf' for confined, 'semi' for semi-confined
        or "leaky" for a leaky top boundary.
    phreatictop : bool
        If true, interpret the first specific storage coefficient as specific
        yield., i.e. it is not multiplied by aquifer thickness.
    tsandhstar : list of tuples
        list containing time and water level pairs for the hstar boundary condition.
    tsandN : list of tuples
        list containing time and infiltration pairs for the infiltration boundary
        condition.
    kzoverkh: float, optional,
        anisotropy factor for vertical resistance, kzoverkh = kz / kh. Default is 1.
    model3d : bool, optional
        Boolean indicating whether model is Model3D-type.
    name : str
        Name of the cross-section inhomogeneity.
    """

    tiny = 1e-12

    def __init__(
        self,
        model,
        x1,
        x2,
        kaq,
        z,
        Haq,
        Hll,
        c,
        Saq,
        Sll,
        poraq,
        porll,
        ltype,
        topboundary,
        phreatictop,
        tsandhstar,
        tsandN,
        kzoverkh=None,
        model3d=False,
        name=None,
    ):
        super().__init__(
            model,
            kaq,
            z,
            Haq,
            Hll,
            c,
            Saq,
            Sll,
            poraq,
            porll,
            ltype,
            topboundary,
            phreatictop,
            kzoverkh=kzoverkh,
            model3d=model3d,
            name=name,
        )
        self.x1 = x1
        self.x2 = x2
        self.tsandhstar = tsandhstar
        self.tsandN = tsandN
        self.inhom_number = self.model.aq.add_inhom(self)
        self.addlinesinks = True  # Set to False not to add line-sinks

    def __repr__(self):
        if self.tsandhstar is not None:
            hstar = " with h*(t)"
        else:
            hstar = ""
        if self.tsandN is not None:
            inf = " with N(t)"
        else:
            inf = ""

        return (
            f"{self.name}: {self.__class__.__name__} "
            + str([self.x1, self.x2])
            + hstar
            + inf
        )

    def is_inside(self, x, _):
        """Check if a point is inside the cross-section.

        Parameters
        ----------
        x : float
            x-coordinate of the point.

        Returns
        -------
        bool
            True if the point is inside the cross-section, False otherwise.
        """
        return (x >= self.x1) and (x < self.x2)

    def initialize(self):
        super().initialize()
        self.create_elements()

    def create_elements(self):
        """Create linesinks to meet the continuity conditions the at the boundaries."""
        if (self.x1 == -np.inf) and (self.x2 == np.inf):
            # no reason to add elements
            pass
        # HeadDiff on right side, FluxDiff on left side
        elif self.x1 == -np.inf:
            xin = self.x2 - self.tiny
            # xoutright = self.x2 + self.tiny
            aqin = self.model.aq.find_aquifer_data(xin, 0)
            # aqoutright = self.model.aq.find_aquifer_data(xoutright, 0)
            if self.addlinesinks:
                HeadDiffLineSink1D(
                    self.model,
                    self.x2,
                    layers=range(self.naq),
                    aq=aqin,
                    label=None,
                )
        elif self.x2 == np.inf:
            xin = self.x1 + self.tiny
            # xoutleft = self.x1 - self.tiny
            aqin = self.model.aq.find_aquifer_data(xin, 0)
            # aqoutleft = self.model.aq.find_aquifer_data(xoutleft, 0)
            if self.addlinesinks:
                FluxDiffLineSink1D(
                    self.model, self.x1, range(self.naq), aq=aqin, label=None
                )
        else:
            xin = 0.5 * (self.x1 + self.x2)
            # xoutleft = self.x1 - self.tiny
            # xoutright = self.x2 + self.tiny
            aqin = self.model.aq.find_aquifer_data(xin, 0)
            # aqleft = self.model.aq.find_aquifer_data(xoutleft, 0)
            # aqright = self.model.aq.find_aquifer_data(xoutright, 0)
            if self.addlinesinks:
                HeadDiffLineSink1D(
                    self.model, self.x2, range(self.naq), label=None, aq=aqin
                )
                FluxDiffLineSink1D(
                    self.model, self.x1, range(self.naq), label=None, aq=aqin
                )
        if self.tsandN is not None:
            assert self.topboundary == "con", Exception(
                "Infiltration can only be applied to a confined aquifer."
            )
            AreaSinkXsection(self.model, self.x1, self.x2, tsandN=self.tsandN)
        if self.tsandhstar is not None:
            assert self.topboundary == "sem", Exception(
                "hstar can only be implemented on top of a semi-confined aquifer."
            )
            HstarXsection(self.model, self.x1, self.x2, tsandhstar=self.tsandhstar)

    def plot(self, ax=None, labels=False, params=False, names=False, fmt=None, **kwargs):
        """Plot the cross-section.

        Parameters
        ----------
        ax : plt.Axes, optional
            Axis to plot the cross-section on. If None, a new axis will be created.
        labels : bool, optional
            If True, add layer-name labels.
        params : bool, optional
            If True, add parameter labels.
        names : bool, optional
            If True, add inhomogeneity names.
        fmt : str, optional
            format string for parameter values, e.g. '.2f' for 2 decimals.
        """
        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(8, 4))

        if "x1" in kwargs:
            x1 = kwargs.pop("x1")
            if np.isfinite(self.x1):
                x1 = max(x1, self.x1)
        elif np.isfinite(self.x1):
            x1 = self.x1
        else:
            x1 = self.x2 - 100.0
        if "x2" in kwargs:
            x2 = kwargs.pop("x2")
            if np.isfinite(self.x2):
                x2 = min(x2, self.x2)
        elif np.isfinite(self.x2):
            x2 = self.x2
        else:
            x2 = self.x1 + 100.0

        if self.x1 > x2 or self.x2 < x1:
            # do nothing, inhom is outside the window
            return ax

        if fmt is None:
            fmt = ""
        ssfmt = ".2e"

        r = x2 - x1
        r0 = x1

        if labels or params:
            lli = 1 if self.topboundary == "con" else 0
            aqi = 0
        else:
            lli = None
            aqi = None

        if names:
            ax.text(
                r0 + 0.5 * r,
                0.95,
                self.name,
                ha="center",
                va="center",
                fontsize=10,
                transform=ax.get_xaxis_transform(),
            )

        for i in range(self.nlayers):
            if self.ltype[i] == "l":
                ax.fill_between(
                    x=[r0, r0 + r],
                    y1=self.z[i + 1],
                    y2=self.z[i],
                    color=[0.8, 0.8, 0.8],
                )
                if labels:
                    ax.text(
                        r0 + 0.5 * r if not params else r0 + 0.25 * r,
                        np.mean(self.z[i : i + 2]),
                        f"leaky layer {lli}",
                        ha="center",
                        va="center",
                    )
                if params:
                    paramtxt = (
                        f"$c$ = {self.c[lli]:{fmt}}, $S_s$ = {self.Sll[lli]:{ssfmt}}"
                    )
                    ax.text(
                        r0 + 0.75 * r if labels else r0 + 0.5 * r,
                        np.mean(self.z[i : i + 2]),
                        paramtxt,
                        ha="center",
                        va="center",
                    )
                if labels or params:
                    lli += 1

            if labels and self.ltype[i] == "a":
                ax.text(
                    r0 + 0.5 * r if not params else r0 + 0.25 * r,
                    np.mean(self.z[i : i + 2]),
                    f"aquifer {aqi}",
                    ha="center",
                    va="center",
                )
            if params and self.ltype[i] == "a":
                if aqi == 0 and i == 0 and self.phreatictop:
                    paramtxt = (
                        f"$k_h$ = {self.kaq[aqi]:{fmt}}, $S$ = {self.Saq[aqi]:{fmt}}"
                    )
                else:
                    paramtxt = (
                        f"$k_h$ = {self.kaq[aqi]:{fmt}}, $S_s$ = {self.Saq[aqi]:{ssfmt}}"
                    )
                ax.text(
                    r0 + 0.75 * r if labels else r0 + 0.5 * r,
                    np.mean(self.z[i : i + 2]),
                    paramtxt,
                    ha="center",
                    va="center",
                )
            if (labels or params) and self.ltype[i] == "a":
                aqi += 1

        for i in range(1, self.nlayers):
            if self.ltype[i] == "a" and self.ltype[i - 1] == "a":
                ax.fill_between(
                    x=[r0, r0 + r],
                    y1=self.z[i],
                    y2=self.z[i],
                    color=[0.8, 0.8, 0.8],
                )

        ax.hlines(self.z[0], xmin=r0, xmax=r0 + r, color="k", lw=0.75)
        ax.hlines(self.z[-1], xmin=r0, xmax=r0 + r, color="k", lw=3.0)
        ax.set_ylabel("elevation")
        return ax


class XsectionMaq(Xsection):
    """Cross-section inhomogeneity consisting of stacked aquifer layers.

    Parameters
    ----------
    model : Model
        Model to add the cross-section to, usually an instance of ModelXsection.
    x1 : float
        x-coordinate of the left boundary of the cross-section.
    x2 : float
        x-coordinate of the right boundary of the cross-section.
    kaq : array
        Hydraulic conductivities of the aquifers.
    z : array
        Elevations of the tops and bottoms of the layers.
    c : array
        Resistance of the leaky layers.
    Saq : array
        Specific storage of the aquifers.
    Sll : array
        Specific storage of the leaky layers.
    poraq : array
        Porosities of the aquifers.
    porll : array
        Porosities of the leaky layers.
    topboundary : str
        Type of top boundary. Can be 'conf' for confined, 'semi' for semi-confined
        or "leaky" for a leaky top boundary.
    phreatictop : bool
        If true, interpret the first specific storage coefficient as specific
        yield., i.e. it is not multiplied by aquifer thickness.
    tsandhstar : list of tuples
        list containing time and water level pairs for the hstar boundary condition.
    tsandN : list of tuples
        list containing time and infiltration pairs for the infiltration boundary
        condition.
    name : str
        Name of the cross-section.
    """

    def __init__(
        self,
        model,
        x1,
        x2,
        kaq=1,
        z=(1, 0),
        c=(),
        Saq=0.001,
        Sll=0,
        poraq=0.3,
        porll=0.3,
        topboundary="conf",
        phreatictop=False,
        tsandhstar=None,
        tsandN=None,
        name=None,
    ):
        kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype = param_maq(
            kaq, z, c, Saq, Sll, poraq, porll, topboundary, phreatictop
        )
        super().__init__(
            model,
            x1,
            x2,
            kaq,
            z,
            Haq,
            Hll,
            c,
            Saq,
            Sll,
            poraq,
            porll,
            ltype,
            topboundary,
            phreatictop,
            tsandhstar,
            tsandN,
            name=name,
        )


class Xsection3D(Xsection):
    """Cross-section inhomogeneity consisting of stacked aquifer layers.

    Vertical resistance is computed from vertical hydraulic conductivity and the
    anisotropy factor.

    Parameters
    ----------
    model : Model
        Model to add the cross-section to, usually an instance of ModelXsection.
    x1 : scalar
        x-coordinate of the left boundary of the cross-section.
    x2 : scalar
        x-coordinate of the right boundary of the cross-section.
    kaq : array
        Hydraulic conductivities of the aquifers.
    z : array
        Elevations of the tops and bottoms of the layers.
    Saq : array
        Specific storage of the aquifers.
    kzoverkh : scalar
        Ratio of vertical hydraulic conductivity to horizontal hydraulic
        conductivity.
    poraq : array
        Porosities of the aquifers.
    topboundary : str
        Type of top boundary. Can be 'conf' for confined, 'semi' for semi-confined
        or "leaky" for a leaky top boundary.
    phreatictop : bool
        If true, interpret the first specific storage coefficient as specific
        yield., i.e. it is not multiplied by aquifer thickness.
    topres : scalar
        Resistance of the top boundary. Only used if topboundary is 'leaky'.
    topthick : scalar
        Thickness of the top boundary. Only used if topboundary is 'leaky'.
    topSll : scalar
        Specific storage of the top boundary. Only used if topboundary is 'leaky'.
    toppor : scalar
        Porosity of the top boundary. Only used if topboundary is 'leaky'.
    tsandhstar : list of tuples
        list containing time and water level pairs for the hstar boundary condition.
    tsandN : list of tuples
        list containing time and infiltration pairs for the infiltration boundary
        condition.
    name : str
        Name of the cross-section.
    """

    def __init__(
        self,
        model,
        x1,
        x2,
        kaq=1,
        z=(4, 3, 2, 1),
        Saq=0.001,
        kzoverkh=0.1,
        poraq=0.3,
        topboundary="conf",
        phreatictop=False,
        topres=0,
        topthick=0,
        topSll=0,
        toppor=0.3,
        tsandhstar=None,
        tsandN=None,
        name=None,
    ):
        kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype, z = param_3d(
            kaq,
            z,
            Saq,
            kzoverkh,
            poraq,
            phreatictop,
            topboundary,
            topres,
            topthick,
            topSll,
            toppor,
        )
        super().__init__(
            model,
            x1,
            x2,
            kaq,
            z,
            Haq,
            Hll,
            c,
            Saq,
            Sll,
            poraq,
            porll,
            ltype,
            topboundary,
            phreatictop,
            tsandhstar,
            tsandN,
            name=name,
        )



# Copyright 2019 Kristopher L. Kuhlman <klkuhlm _at_ sandia _dot_ gov>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numba
import numpy as np


@numba.njit(nogil=True, cache=True)
def invlap(t, tmax, fp, M, alpha=1e-10, tol=1e-9):
    """Inverse Laplace tansform with algorithm of De Hoog, Knight and Stokes.

    Parameters
    ----------
    t : array
        times for which inverse is computed
    tmax : float
        maximum time
    fp : complex array
        Laplace transformed solution
    M : integer
        number of terms (number of values of fp)
    alpha:  is the real part of the rightmost pole or singularity, which
            is chosen based on the desired accuracy (assuming the rightmost
            singularity is 0), and tol=10α is the desired tolerance

    Returns
    -------
    result : array
        time domain solution for specified times

    Reference
    ---------

    de Hoog, F., J. Knight, A. Stokes (1982). An improved method for
    numerical inversion of Laplace transforms. SIAM Journal of Scientific
    and Statistical Computing 3:357-366, http://dx.doi.org/10.1137/0903022

    https://bitbucket.org/klkuhlm/invlap/src/default/invlap.py
    """
    # return zeros if all fp equal to zero
    if np.all(np.abs(fp) == 0):
        return np.zeros(len(t))

    NP = 2 * M + 1
    scale = 2.0
    T = scale * tmax
    Nt = len(t)
    #
    # tol = alpha * 10.0
    gamma = alpha - np.log(tol) / (scale * T)

    # would it be useful to try re-using
    # space between e&q and A&B?
    # Kris programmed this as np.complex64
    e = np.empty((2 * M + 1, M + 1), dtype=np.complex128)
    q = np.empty((2 * M, M), dtype=np.complex128)
    d = np.empty(2 * M + 1, dtype=np.complex128)
    A = np.empty((2 * M + 2, Nt), dtype=np.complex128)
    B = np.empty((2 * M + 2, Nt), dtype=np.complex128)

    # initialize Q-D table
    e[:, 0] = 0.0
    q[0, 0] = fp[1] / (fp[0] / 2.0)
    for i in range(1, 2 * M):
        q[i, 0] = fp[i + 1] / fp[i]

    # rhombus rule for filling triangular Q-D table (e & q)
    for r in range(1, M + 1):
        # start with e, column 1, 0:2*M-2
        mr = 2 * (M - r) + 1
        e[0:mr, r] = q[1 : mr + 1, r - 1] - q[0:mr, r - 1] + e[1 : mr + 1, r - 1]
        if not r == M:
            rq = r + 1
            mr = 2 * (M - rq) + 2
            for i in range(mr):
                q[i, rq - 1] = q[i + 1, rq - 2] * e[i + 1, rq - 1] / e[i, rq - 1]

    # build up continued fraction coefficients (d)
    d[0] = fp[0] / 2.0
    for r in range(1, M + 1):
        d[2 * r - 1] = -q[0, r - 1]  # even terms
        d[2 * r] = -e[0, r]  # odd terms

    # seed A and B for recurrence
    A[0] = 0.0
    A[1] = d[0]
    B[0:2] = 1.0 + 0j

    # base of the power series
    z = np.exp(1j * np.pi * t / T)

    # coefficients of Pade approximation (A & B)
    # using recurrence for all but last term
    for i in range(1, 2 * M):
        A[i + 1] = A[i] + d[i] * A[i - 1] * z
        B[i + 1] = B[i] + d[i] * B[i - 1] * z

    # "improved remainder" to continued fraction
    brem = (1.0 + (d[2 * M - 1] - d[2 * M]) * z) / 2.0
    rem = -brem * (1.0 - np.sqrt(1.0 + d[2 * M] * z / brem**2))

    # last term of recurrence using new remainder
    A[NP] = A[2 * M] + rem * A[2 * M - 1]
    B[NP] = B[2 * M] + rem * B[2 * M - 1]

    # diagonal Pade approximation
    # F=A/B represents accelerated trapezoid rule
    result = np.exp(gamma * t) / T * (A[NP] / B[NP]).real
    # print('results:', result)
    # print('A', A)
    # print('B', B)

    return result


@numba.njit(nogil=True, cache=True)
def compute_laplace_parameters_numba(tmax, M=20, alpha=1e-10, tol=1e-9):
    # 2*M+1 terms in approximation
    # desired tolerance (here simply related to alpha)
    # tol = alpha * 10.0
    nump = 2 * M + 1  # number of terms in approximation
    # scaling factor (likely tune-able, but 2 is typical)
    scale = 2.0
    T = scale * tmax
    gamma = alpha - np.log(tol) / (scale * T)
    p = gamma + 1j * np.pi * np.arange(nump) / T
    return p


def invlaptest():
    p = compute_laplace_parameters_numba(tmax=10, alpha=1e-10)
    fp = 1 / (p + 1) ** 2
    t = np.arange(1.0, 10)
    ft = invlap(t, 10, fp, 20, alpha=1e-10)
    print("approximate from invlap:", ft)
    print("exact:", t * np.exp(-t))


@numba.njit(nogil=True, cache=True)
def invlapcomp(time, pot, npint, M, tintervals, enumber, etstart, ebc, nlayers):
    """Compute time domain solution for given laplace domain solution.

    Parameters
    ----------
    time : array, must be ordered
           times for which time domain solution is computed, must start at 0
    pot : array of laplace domain solution conform the ttim shape
    npint : int
        number of p values per interval (=2M + 1)
    M : int
        order of the approximation
    tintervals : time intervals
    enumber : array with number of element
    etstart : array with starting time of bc in element
    ebc : array with boundary condition value of element
    nlayers : integer or None (default)
        number of layers

    Notes
    -----
    enumber, etstart, and ebc are used because numba cannot deal with a list
    of arrays of different lengths (makes sense, actually)

    Returns
    -------
    pot[naq, ntimes] if layers=None,
    otherwise pot[len(layers) ,ntimes]
    t must be ordered
    """
    print_tmin_warning = True  # set to False if warning is printed once
    print_tmax_warning = True
    nelements, naq, npval = pot.shape
    nint = len(tintervals) - 1
    rv = np.zeros((nlayers, len(time)))
    #
    # assuming that first time of all bc's is 0
    for j in range(len(enumber)):
        t = time - etstart[j]
        it = 0
        if t[0] <= 0:  # there are times before start of bc
            if t[-1] <= 0:  # all times before start of bc, also for len(t)=1
                continue
            else:
                # no effect for any t <= 0
                it = np.argmax(t > 0)  # find_first
        if t[it] < tintervals[0]:  # there are times before first interval
            if print_tmin_warning:
                print("Warning, some of the times are smaller than tmin after")
                print("a change in boundary condition. nans are substituted")
                print_tmin_warning = False
            if t[-1] < tintervals[0]:  # all times before first interval
                itnew = len(t)
            else:
                itnew = np.argmax(t >= tintervals[0])  # find_first
            rv[:, it:itnew] = np.nan
            it = itnew
        for n in range(nint):
            if n == 0:
                tp = t[(t >= tintervals[n]) & (t <= tintervals[n + 1])]
            else:
                tp = t[(t > tintervals[n]) & (t <= tintervals[n + 1])]
            nt = len(tp)
            # if nt > 0:  # if all zero, don't do the inv transform
            if nt == 0:
                continue
            for i in range(nlayers):
                # I used to check the first value only, but got to check if
                # none of the values are zero
                if not np.any(pot[enumber[j], i, n * npint : (n + 1) * npint] == 0):
                    rv[i, it : it + nt] += ebc[j] * invlap(
                        tp,
                        tintervals[n + 1],
                        pot[enumber[j], i, n * npint : (n + 1) * npint],
                        M,
                    )
            it = it + nt
        if it < len(t):  # there are times above tintervals[-1]
            if print_tmax_warning:
                print("Warning, some of the times are larger than tmax after")
                print("a change in boundary condition. nans are substituted")
                print_tmax_warning = False
            rv[:, it:] = np.nan
    return rv


# The following general function is currently not used but very useful
# to do numerical Laplace transformation
@numba.njit(nogil=True, cache=True)
def invlapgen(time, pot, M, tintervals, tstart, ebc):
    """Compute time domain solution for given Laplace domain solution.

    Parameters
    ----------
    time : array, must be ordered
           times for which time domain solution is computed, must start at 0
    pot : 1D array of laplace domain solution of length nint * (2M + 1)
    M : int order of the approximation (i.e., (2M + 1) p values per interval)
    tintervals : 1D array of time intervals, length nint + 1
    tstart : 1D array with starting times of bc in element, length nstart
    ebc : 1D array with change in boundary condition value, length nstart

    Notes
    -----
    ebc is the difference with the previous value

    Returns
    -------
    rv[ntimes]
    """
    print_tmin_warning = True  # set to False if warning is printed once
    print_tmax_warning = True
    npint = 2 * M + 1
    nint = len(tintervals) - 1
    rv = np.zeros(len(time))
    #
    # assuming that first time of all bc's is 0
    for j in range(len(tstart)):
        t = time - tstart[j]
        it = 0
        if t[0] <= 0:  # there are times before start of bc
            if t[-1] <= 0:  # all times before start of bc, also for len(t)=1
                continue
            else:
                # no effect for any t <= 0
                it = np.argmax(t > 0)  # find_first
        if t[it] < tintervals[0]:  # there are times before first interval
            if print_tmin_warning:
                print("Warning, some of the times are smaller than tmin after")
                print("a change in boundary condition. nans are substituted")
                print_tmin_warning = False
            if t[-1] < tintervals[0]:  # all times before first interval
                itnew = len(t)
            else:
                itnew = np.argmax(t >= tintervals[0])  # find_first
            rv[it:itnew] = np.nan
            it = itnew
        for n in range(nint):
            if n == 0:
                tp = t[(t >= tintervals[n]) & (t <= tintervals[n + 1])]
            else:
                tp = t[(t > tintervals[n]) & (t <= tintervals[n + 1])]
            nt = len(tp)
            # if nt > 0:  # if all zero, don't do the inv transform
            if nt == 0:
                continue
            # I used to check the first value only, but got to check if
            # none of the values are zero
            if not np.any(pot[n * npint : (n + 1) * npint] == 0):
                rv[it : it + nt] += ebc[j] * invlap(
                    tp, tintervals[n + 1], pot[n * npint : (n + 1) * npint], M
                )
            it = it + nt
        if it < len(t):  # there are times above tintervals[-1]
            if print_tmax_warning:
                print("Warning, some of the times are larger than tmax after")
                print("a change in boundary condition. nans are substituted")
                print_tmax_warning = False
            rv[it:] = np.nan
    return rv



# flake8: noqa
import numpy as np
from scipy.special import iv  # Needed for K1 in Well class, and in CircInhom
from scipy.special import kv

from .aquifer import AquiferData
from .element import Element
from .equation import InhomEquation


class CircInhomData(AquiferData):
    def __init__(
        self,
        model,
        x0=0,
        y0=0,
        R=1,
        kaq=[1],
        Haq=[1],
        c=[1],
        Saq=[0.1],
        Sll=[0.1],
        topboundary="imp",
    ):
        AquiferData.__init__(
            self, model, kaq, Haq, Hll, c, Saq, Sll, topboundary, phreatictop
        )
        self.x0 = float(x0)
        self.y0 = float(y0)
        self.R = float(R)
        self.Rsq = self.R**2
        self.area = np.pi * self.Rsq
        self.model.addInhom(self)

    def is_inside(self, x, y):
        rv = False
        if (x - self.x0) ** 2 + (y - self.y0) ** 2 < self.Rsq:
            rv = True
        return rv


class CircInhomDataMaq(CircInhomData):
    def __init__(
        self,
        model,
        x0=0,
        y0=0,
        R=1,
        kaq=[1],
        z=[1, 0],
        c=[],
        Saq=[0.001],
        Sll=[0],
        topboundary="imp",
        phreatictop=False,
    ):
        kaq, Haq, Hll, c, Saq, Sll = param_maq(
            kaq, z, c, Saq, Sll, topboundary, phreatictop
        )
        CircInhomData.__init__(
            self, model, x0, y0, R, kaq, Haq, c, Saq, Sll, topboundary, phreatictop
        )


class CircInhomData3D(CircInhomData):
    def __init__(
        self,
        model,
        x0=0,
        y0=0,
        R=1,
        kaq=1,
        z=[4, 3, 2, 1],
        Saq=[0.3, 0.001, 0.001],
        kzoverkh=0.1,
        phreatictop=True,
        topboundary="conf",
        topres=0,
        topthick=0,
        topSll=0,
    ):
        kaq, Haq, Hll, c, Saq, Sll = param_3d(
            kaq, z, Saq, kzoverkh, phreatictop, topboundary, topres, topthick, topSll
        )
        CircInhomData.__init__(self, model, x0, y0, R, kaq, Haq, c, Saq, Sll, "imp")


class BesselRatioApprox:
    # Never fully debugged
    def __init__(self, Norder, Nterms):
        self.Norder = Norder + 1
        self.Nterms = Nterms + 1
        self.krange = np.arange(self.Nterms)
        self.minonek = (-np.ones(self.Nterms)) ** self.krange
        self.hankeltot = np.ones((self.Norder, 2 * self.Nterms), dtype=float)
        self.muk = np.ones((self.Norder, self.Nterms), dtype=float)
        self.nuk = np.ones((self.Norder, self.Nterms), dtype=float)
        for n in range(self.Norder):
            mu = 4.0 * n**2
            for k in range(1, self.Nterms):
                self.hankeltot[n, k] = (
                    self.hankeltot[n, k - 1] * (mu - (2 * k - 1) ** 2) / (4.0 * k)
                )
            for k in range(self.Nterms):
                self.muk[n, k] = (4.0 * n**2 + 16.0 * k**2 - 1.0) / (
                    4.0 * n**2 - (4.0 * k - 1.0) ** 2
                )
                self.nuk[n, k] = (4.0 * n**2 + 4.0 * (2.0 * k + 1.0) ** 2 - 1.0) / (
                    4.0 * n**2 - (4.0 * k + 1.0) ** 2
                )
        self.hankelnk = self.hankeltot[:, : self.Nterms]
        self.hankeln2k = self.hankeltot[:, ::2]
        self.hankeln2kp1 = self.hankeltot[:, 1::2]

    def ivratio(self, rho, R, lab):
        lab = np.atleast_1d(lab)
        rv = np.empty((self.Norder, len(lab)), dtype=complex)
        for k in range(len(lab)):
            top = np.sum(
                self.minonek * self.hankelnk / (2.0 * rho / lab[k]) ** self.krange, 1
            )
            bot = np.sum(
                self.minonek * self.hankelnk / (2.0 * R / lab[k]) ** self.krange, 1
            )
            rv[:, k] = top / bot * np.sqrt(float(R) / rho) * np.exp((rho - R) / lab[k])
        return rv

    def kvratio(self, rho, R, lab):
        lab = np.atleast_1d(lab)
        rv = np.empty((self.Norder, len(lab)), dtype=complex)
        for k in range(len(lab)):
            top = np.sum(self.hankelnk / (2.0 * rho / lab[k]) ** self.krange, 1)
            bot = np.sum(self.hankelnk / (2.0 * R / lab[k]) ** self.krange, 1)
            rv[:, k] = top / bot * np.sqrt(float(R) / rho) * np.exp((R - rho) / lab[k])
        return rv

    def ivratiop(self, rho, R, lab):
        lab = np.atleast_1d(lab)
        rv = np.empty((self.Norder, len(lab)), dtype=complex)
        for k in range(len(lab)):
            top = np.sum(
                self.muk * self.hankeln2k / (2.0 * rho / lab[k]) ** (2 * self.krange), 1
            ) - np.sum(
                self.nuk
                * self.hankeln2kp1
                / (2.0 * rho / lab[k]) ** (2 * self.krange + 1),
                1,
            )
            bot = np.sum(
                self.minonek * self.hankelnk / (2.0 * R / lab[k]) ** self.krange, 1
            )
            rv[:, k] = top / bot * np.sqrt(float(R) / rho) * np.exp((rho - R) / lab[k])
        return rv

    def kvratiop(self, rho, R, lab):
        lab = np.atleast_1d(lab)
        rv = np.empty((self.Norder, len(lab)), dtype=complex)
        for k in range(len(lab)):
            top = np.sum(
                self.muk * self.hankeln2k / (2.0 * rho / lab[k]) ** (2 * self.krange), 1
            ) + np.sum(
                self.nuk
                * self.hankeln2kp1
                / (2.0 * rho / lab[k]) ** (2 * self.krange + 1),
                1,
            )
            bot = np.sum(self.hankelnk / (2.0 * R / lab[k]) ** self.krange, 1)
            rv[:, k] = -top / bot * np.sqrt(float(R) / rho) * np.exp((R - rho) / lab[k])
        return rv


class CircInhomRadial(Element, InhomEquation):
    def __init__(self, model, x0=0, y0=0, R=1.0, label=None):
        Element.__init__(
            self,
            model,
            nparam=2 * model.aq.naq,
            nunknowns=2 * model.aq.naq,
            layers=range(model.aq.naq),
            type="z",
            name="CircInhom",
            label=label,
        )
        self.x0 = float(x0)
        self.y0 = float(y0)
        self.R = float(R)
        self.model.addElement(self)
        self.approx = BesselRatioApprox(0, 2)

    def __repr__(self):
        return self.name + " at " + str((self.x0, self.y0))

    def initialize(self):
        self.xc = np.array([self.x0 + self.R])
        self.yc = np.array([self.y0])
        self.thetacp = np.zeros(1)
        self.ncp = 1
        self.aqin = self.model.aq.findAquiferData(self.x0 + (1 - 1e-8) * self.R, self.y0)
        assert self.aqin.R == self.R, (
            "Radius of CircInhom and CircInhomData must be equal"
        )
        self.aqout = self.model.aq.find_aquifer_data(
            self.x0 + (1 + 1e-8) * self.R, self.y0
        )
        self.setbc()
        self.facin = np.ones_like(self.aqin.lab2)
        self.facout = np.ones_like(self.aqout.lab2)
        # To keep track which circles are small
        self.circ_in_small = np.ones((self.aqin.naq, self.model.nin), dtype="i")
        self.circ_out_small = np.ones((self.aqout.naq, self.model.nin), dtype="i")
        self.Rbig = 700
        # for i in range(self.aqin.Naq):
        #    for j in range(self.model.Nin):
        #        assert self.R / abs(self.aqin.lab2[i,j,0]) < self.Rbig, 'TTim input error, Radius too big'
        #        assert self.R / abs(self.aqout.lab2[i,j,0]) < self.Rbig, 'TTim input error, Radius too big'
        #        if self.R / abs(self.aqin.lab2[i,j,0]) < self.Rbig:
        #            self.circ_in_small[i,j] = 1
        #            self.facin[i,j,:] = 1.0 / iv(0, self.R / self.aqin.lab2[i,j,:])
        #        if self.R / abs(self.aqout.lab2[i,j,0]) < self.Rbig:
        #            self.circ_out_small[i,j] = 1
        #            self.facout[i,j,:] = 1.0 / kv(0, self.R / self.aqout.lab2[i,j,:])
        # for i in range(self.aqin.Naq):
        #    for j in range(self.model.Nin):
        #        assert self.R / abs(self.aqin.lab2[i,j,0]) < 900, 'radius too large compared to aqin lab2[i,j,0] '+str((i,j))
        #        assert self.R / abs(self.aqout.lab2[i,j,0]) < 900, 'radius too large compared to aqin lab2[i,j,0] '+str((i,j))
        # self.facin = 1.0 / iv(0, self.R / self.aqin.lab2)
        # self.facout = 1.0 / kv(0, self.R / self.aqout.lab2)
        self.parameters = np.zeros(
            (self.model.Ngvbc, self.Nparam, self.model.Np), dtype=complex
        )

    def potinf(self, x, y, aq=None):
        """Can be called with only one x,y value."""
        if aq is None:
            aq = self.model.aq.findAquiferData(x, y)
        rv = np.zeros(
            (self.nparam, aq.naq, self.model.nin, self.model.npin), dtype=complex
        )
        if aq == self.aqin:
            r = np.sqrt((x - self.x0) ** 2 + (y - self.y0) ** 2)
            for i in range(self.aqin.Naq):
                for j in range(self.model.Nin):
                    if abs(r - self.R) / abs(self.aqin.lab2[i, j, 0]) < self.Rzero:
                        if self.circ_in_small[i, j]:
                            rv[i, i, j, :] = self.facin[i, j, :] * iv(
                                0, r / self.aqin.lab2[i, j, :]
                            )
                        else:
                            print("using approx")
                            rv[i, i, j, :] = self.approx.ivratio(
                                r, self.R, self.aqin.lab2[i, j, :]
                            )
        if aq == self.aqout:
            r = np.sqrt((x - self.x0) ** 2 + (y - self.y0) ** 2)
            for i in range(self.aqout.Naq):
                for j in range(self.model.Nin):
                    if abs(r - self.R) / abs(self.aqout.lab2[i, j, 0]) < self.Rzero:
                        if self.circ_out_small[i, j]:
                            rv[self.aqin.Naq + i, i, j, :] = self.facin[i, j, :] * kv(
                                0, r / self.aqout.lab2[i, j, :]
                            )
                        else:
                            print("using approx")
                            rv[self.aqin.Naq + i, i, j, :] = self.approx.kvratio(
                                r, self.R, self.aqout.lab2[i, j, :]
                            )
        rv.shape = (self.Nparam, aq.Naq, self.model.Np)
        return rv

    def disinf(self, x, y, aq=None):
        """Can be called with only one x,y value."""
        if aq is None:
            aq = self.model.aq.findAquiferData(x, y)
        qx = np.zeros((self.nparam, aq.naq, self.model.np), dtype=complex)
        qy = np.zeros((self.nparam, aq.naq, self.model.np), dtype=complex)
        if aq == self.aqin:
            qr = np.zeros(
                (self.nparam, aq.naq, self.model.nin, self.model.npin), dtype=complex
            )
            r = np.sqrt((x - self.x0) ** 2 + (y - self.y0) ** 2)
            if r < 1e-20:
                r = 1e-20  # As we divide by that on the return
            for i in range(self.aqin.Naq):
                for j in range(self.model.Nin):
                    if abs(r - self.R) / abs(self.aqin.lab2[i, j, 0]) < self.Rzero:
                        if self.circ_in_small[i, j]:
                            qr[i, i, j, :] = (
                                -self.facin[i, j, :]
                                * iv(1, r / self.aqin.lab2[i, j, :])
                                / self.aqin.lab2[i, j, :]
                            )
                        else:
                            qr[i, i, j, :] = (
                                -self.approx.ivratiop(r, self.R, self.aqin.lab2[i, j, :])
                                / self.aqin.lab2[i, j, :]
                            )
            qr.shape = (self.nparam, aq.naq, self.model.np)
            qx[:] = qr * (x - self.x0) / r
            qy[:] = qr * (y - self.y0) / r
        if aq == self.aqout:
            qr = np.zeros(
                (self.Nparam, aq.Naq, self.model.Nin, self.model.Npin), dtype=complex
            )
            r = np.sqrt((x - self.x0) ** 2 + (y - self.y0) ** 2)
            for i in range(self.aqout.Naq):
                for j in range(self.model.Nin):
                    if abs(r - self.R) / abs(self.aqout.lab2[i, j, 0]) < self.Rzero:
                        if self.circ_out_small[i, j]:
                            qr[self.aqin.Naq + i, i, j, :] = (
                                self.facin[i, j, :]
                                * kv(1, r / self.aqout.lab2[i, j, :])
                                / self.aqout.lab2[i, j, :]
                            )
                        else:
                            qr[self.aqin.Naq + i, i, j, :] = (
                                self.approx.kvratiop(r, self.R, self.aqout.lab2[i, j, :])
                                / self.aqout.lab2[i, j, :]
                            )
            qr.shape = (self.Nparam, aq.Naq, self.model.Np)
            qx[:] = qr * (x - self.x0) / r
            qy[:] = qr * (y - self.y0) / r
        return qx, qy

    def layout(self):
        alpha = np.linspace(0, 2 * np.pi, 100)
        return (
            "line",
            self.x0 + self.R * np.cos(alpha),
            self.y0 + self.R * np.sin(alpha),
        )


# class CircInhom(Element,InhomEquation):
#     def __init__(self,model,x0=0,y0=0,R=1.0,order=0,label=None,test=False):
#         Element.__init__(self, model, Nparam=2*model.aq.Naq*(2*order+1), Nunknowns=2*model.aq.Naq*(2*order+1), layers=range(model.aq.Naq), type='z', name='CircInhom', label=label)
#         self.x0 = float(x0); self.y0 = float(y0); self.R = float(R)
#         self.order = order
#         self.approx = BesselRatioApprox(0,3)
#         self.test=test
#         self.model.addElement(self)
#     def __repr__(self):
#         return self.name + ' at ' + str((self.x0,self.y0))
#     def initialize(self):
#         self.Ncp = 2*self.order + 1
#         self.thetacp = np.arange(0,2*np.pi,(2*np.pi)/self.Ncp)
#         self.xc = self.x0 + self.R * np.cos( self.thetacp )
#         self.yc = self.y0 + self.R * np.sin( self.thetacp )
#         self.aqin = self.model.aq.findAquiferData(self.x0 + (1-1e-10)*self.R,self.y0)
#         self.aqout = self.model.aq.findAquiferData(self.x0+(1.0+1e-8)*self.R,self.y0)
#         assert self.aqin.Naq == self.aqout.Naq, 'TTim input error: Number of layers needs to be the same inside and outside circular inhomogeneity'
#         # Now that aqin is known, check that radii of circles are the same
#         assert self.aqin.R == self.R, 'TTim Input Error: Radius of CircInhom and CircInhomData must be equal'
#         self.setbc()
#         self.facin = np.zeros((self.order+1,self.aqin.Naq,self.model.Nin,self.model.Npin),dtype='D')
#         self.facout = np.zeros((self.order+1,self.aqin.Naq,self.model.Nin,self.model.Npin),dtype='D')
#         self.circ_in_small = np.zeros((self.aqin.Naq,self.model.Nin),dtype='i') # To keep track which circles are small
#         self.circ_out_small = np.zeros((self.aqout.Naq,self.model.Nin),dtype='i')
#         self.besapprox = BesselRatioApprox(self.order,2) # Nterms = 2 is probably enough
#         self.Rbig = 200
#         for i in range(self.aqin.Naq):
#             for j in range(self.model.Nin):
#                 # When the circle is too big, an assertion is thrown. In the future, the approximation of the ratio of bessel functions needs to be completed
#                 # For now, the logic is there, but not used
#                 if self.test:
#                     print('inside  relative radius: ',self.R / abs(self.aqin.lab2[i,j,0]))
#                     print('outside relative radius: ',self.R / abs(self.aqout.lab2[i,j,0]))
#                 #assert self.R / abs(self.aqin.lab2[i,j,0]) < self.Rbig, 'TTim input error, Radius too big'
#                 #assert self.R / abs(self.aqout.lab2[i,j,0]) < self.Rbig, 'TTim input error, Radius too big'
#                 if self.R / abs(self.aqin.lab2[i,j,0]) < self.Rbig:
#                     self.circ_in_small[i,j] = 1
#                     for n in range(self.order+1):
#                         self.facin[n,i,j,:] = 1.0 / iv(n, self.R / self.aqin.lab2[i,j,:])
#                 if self.R / abs(self.aqout.lab2[i,j,0]) < self.Rbig:
#                     self.circ_out_small[i,j] = 1
#                     for n in range(self.order+1):
#                         self.facout[n,i,j,:] = 1.0 / kv(n, self.R / self.aqout.lab2[i,j,:])
#         self.parameters = np.zeros( (self.model.Ngvbc, self.Nparam, self.model.Np), 'D' )
#     def potinf(self,x,y,aq=None):
#         '''Can be called with only one x,y value'''
#         if aq is None: aq = self.model.aq.findAquiferData( x, y )
#         rv = np.zeros((2*aq.Naq,1+2*self.order,aq.Naq,self.model.Nin,self.model.Npin),'D')
#         if aq == self.aqin:
#             r = np.sqrt( (x-self.x0)**2 + (y-self.y0)**2 )
#             alpha = np.arctan2(y-self.y0, x-self.x0)
#             for i in range(self.aqin.Naq):
#                 for j in range(self.model.Nin):
#                     if abs(r-self.R) / abs(self.aqin.lab2[i,j,0]) < self.Rzero:
#                         if self.circ_in_small[i,j]:
#                             pot = np.zeros((self.model.Npin),'D')
#                             rv[i,0,i,j,:] = iv( 0, r / self.aqin.lab2[i,j,:] ) * self.facin[0,i,j,:]
#                             for n in range(1,self.order+1):
#                                 pot[:] = iv( n, r / self.aqin.lab2[i,j,:] ) * self.facin[n,i,j,:]
#                                 rv[i,2*n-1,i,j,:] = pot * np.cos(n*alpha)
#                                 rv[i,2*n  ,i,j,:] = pot * np.sin(n*alpha)
#                         else:
#                             pot = self.besapprox.ivratio(r,self.R,self.aqin.lab2[i,j,:])
#                             rv[i,0,i,j,:] = pot[0]
#                             for n in range(1,self.order+1):
#                                 rv[i,2*n-1,i,j,:] = pot[n] * np.cos(n*alpha)
#                                 rv[i,2*n  ,i,j,:] = pot[n] * np.sin(n*alpha)
#         if aq == self.aqout:
#             r = np.sqrt( (x-self.x0)**2 + (y-self.y0)**2 )
#             alpha = np.arctan2(y-self.y0, x-self.x0)
#             for i in range(self.aqout.Naq):
#                 for j in range(self.model.Nin):
#                     if abs(r-self.R) / abs(self.aqout.lab2[i,j,0]) < self.Rzero:
#                         if self.circ_out_small[i,j]:
#                             pot = np.zeros((self.model.Npin),'D')
#                             rv[aq.Naq+i,0,i,j,:] = kv( 0, r / self.aqout.lab2[i,j,:] ) * self.facout[0,i,j,:]
#                             for n in range(1,self.order+1):
#                                 pot[:] = kv( n, r / self.aqout.lab2[i,j,:] ) * self.facout[n,i,j,:]
#                                 rv[aq.Naq+i,2*n-1,i,j,:] = pot * np.cos(n*alpha)
#                                 rv[aq.Naq+i,2*n  ,i,j,:] = pot * np.sin(n*alpha)
#                         else:
#                             pot = self.besapprox.kvratio(r,self.R,self.aqout.lab2[i,j,:])
#                             rv[aq.Naq+i,0,i,j,:] = pot[0]
#                             for n in range(1,self.order+1):
#                                 rv[aq.Naq+i,2*n-1,i,j,:] = pot[n] * np.cos(n*alpha)
#                                 rv[aq.Naq+i,2*n  ,i,j,:] = pot[n] * np.sin(n*alpha)
#         rv.shape = (self.Nparam,aq.Naq,self.model.Np)
#         return rv
#     def disinf(self,x,y,aq=None):
#         '''Can be called with only one x,y value'''
#         if aq is None: aq = self.model.aq.findAquiferData( x, y )
#         qx = np.zeros((self.Nparam,aq.Naq,self.model.Np),'D')
#         qy = np.zeros((self.Nparam,aq.Naq,self.model.Np),'D')
#         if aq == self.aqin:
#             r = np.sqrt( (x-self.x0)**2 + (y-self.y0)**2 )
#             alpha = np.arctan2(y-self.y0, x-self.x0)
#             qr = np.zeros((aq.Naq,1+2*self.order,aq.Naq,self.model.Nin,self.model.Npin),'D')
#             qt = np.zeros((aq.Naq,1+2*self.order,aq.Naq,self.model.Nin,self.model.Npin),'D')
#             if r < 1e-20: r = 1e-20  # As we divide by that on the return
#             for i in range(self.aqin.Naq):
#                 for j in range(self.model.Nin):
#                     if abs(r-self.R) / abs(self.aqin.lab2[i,j,0]) < self.Rzero:
#                         if self.circ_in_small[i,j]:
#                             pot = np.zeros((self.order+2,self.model.Npin),'D')
#                             for n in range(self.order+2):
#                                 pot[n] = iv( n, r / self.aqin.lab2[i,j,:] )
#                             qr[i,0,i,j,:] = -pot[1] / self.aqin.lab2[i,j,:] * self.facin[0,i,j,:]
#                             for n in range(1,self.order+1):
#                                 qr[i,2*n-1,i,j,:] = -(pot[n-1] + pot[n+1]) / 2 / self.aqin.lab2[i,j,:] * np.cos(n*alpha) * self.facin[n,i,j,:]
#                                 qr[i,2*n  ,i,j,:] = -(pot[n-1] + pot[n+1]) / 2 / self.aqin.lab2[i,j,:] * np.sin(n*alpha) * self.facin[n,i,j,:]
#                                 qt[i,2*n-1,i,j,:] =   pot[n] * np.sin(n*alpha) * n / r * self.facin[n,i,j,:]
#                                 qt[i,2*n  ,i,j,:] =  -pot[n] * np.cos(n*alpha) * n / r * self.facin[n,i,j,:]
#                         else:
#                             pot  = self.besapprox.ivratio(r,self.R,self.aqin.lab2[i,j,:])
#                             potp = self.besapprox.ivratiop(r,self.R,self.aqin.lab2[i,j,:])
#                             qr[i,0,i,j,:] = -potp[0] / self.aqin.lab2[i,j,:]
#                             for n in range(1,self.order+1):
#                                 qr[i,2*n-1,i,j,:] = -potp[n] / self.aqin.lab2[i,j,:] * np.cos(n*alpha)
#                                 qr[i,2*n  ,i,j,:] = -potp[n] / 2 / self.aqin.lab2[i,j,:] * np.sin(n*alpha)
#                                 qt[i,2*n-1,i,j,:] =  pot[n] * np.sin(n*alpha) * n / r
#                                 qt[i,2*n  ,i,j,:] = -pot[n] * np.cos(n*alpha) * n / r
#             qr.shape = (self.Nparam/2,aq.Naq,self.model.Np)
#             qt.shape = (self.Nparam/2,aq.Naq,self.model.Np)
#             qx[:self.Nparam/2,:,:] = qr * np.cos(alpha) - qt * np.sin(alpha);
#             qy[:self.Nparam/2,:,:] = qr * np.sin(alpha) + qt * np.cos(alpha);
#         if aq == self.aqout:
#             r = np.sqrt( (x-self.x0)**2 + (y-self.y0)**2 )
#             alpha = np.arctan2(y-self.y0, x-self.x0)
#             qr = np.zeros((aq.Naq,1+2*self.order,aq.Naq,self.model.Nin,self.model.Npin),'D')
#             qt = np.zeros((aq.Naq,1+2*self.order,aq.Naq,self.model.Nin,self.model.Npin),'D')
#             if r < 1e-20: r = 1e-20  # As we divide by that on the return
#             for i in range(self.aqout.Naq):
#                 for j in range(self.model.Nin):
#                     if abs(r-self.R) / abs(self.aqout.lab2[i,j,0]) < self.Rzero:
#                         if self.circ_out_small[i,j]:
#                             pot = np.zeros((self.order+2,self.model.Npin),'D')
#                             for n in range(self.order+2):
#                                 pot[n] = kv( n, r / self.aqout.lab2[i,j,:] )
#                             qr[i,0,i,j,:] = pot[1] / self.aqout.lab2[i,j,:] * self.facout[0,i,j,:]
#                             for n in range(1,self.order+1):
#                                 qr[i,2*n-1,i,j,:] = (pot[n-1] + pot[n+1]) / 2 / self.aqout.lab2[i,j,:] * np.cos(n*alpha) * self.facout[n,i,j,:]
#                                 qr[i,2*n  ,i,j,:] = (pot[n-1] + pot[n+1]) / 2 / self.aqout.lab2[i,j,:] * np.sin(n*alpha) * self.facout[n,i,j,:]
#                                 qt[i,2*n-1,i,j,:] =   pot[n] * np.sin(n*alpha) * n / r * self.facout[n,i,j,:]
#                                 qt[i,2*n  ,i,j,:] =  -pot[n] * np.cos(n*alpha) * n / r * self.facout[n,i,j,:]
#                         else:
#                             pot  = self.besapprox.kvratio(r,self.R,self.aqout.lab2[i,j,:])
#                             potp = self.besapprox.kvratiop(r,self.R,self.aqout.lab2[i,j,:])
#                             qr[i,0,i,j,:] = -potp[0] / self.aqout.lab2[i,j,:]
#                             for n in range(1,self.order+1):
#                                 qr[i,2*n-1,i,j,:] = -potp[n] / self.aqout.lab2[i,j,:] * np.cos(n*alpha)
#                                 qr[i,2*n  ,i,j,:] = -potp[n] / self.aqout.lab2[i,j,:] * np.sin(n*alpha)
#                                 qt[i,2*n-1,i,j,:] =  pot[n] * np.sin(n*alpha) * n / r
#                                 qt[i,2*n  ,i,j,:] = -pot[n] * np.cos(n*alpha) * n / r
#             qr.shape = (self.Nparam/2,aq.Naq,self.model.Np)
#             qt.shape = (self.Nparam/2,aq.Naq,self.model.Np)
#             qx[self.Nparam/2:,:,:] = qr * np.cos(alpha) - qt * np.sin(alpha);
#             qy[self.Nparam/2:,:,:] = qr * np.sin(alpha) + qt * np.cos(alpha);
#         return qx,qy
#     def layout(self):
#         return 'line', self.x0 + self.R * np.cos(np.linspace(0,2*np.pi,100)), self.y0 + self.R * np.sin(np.linspace(0,2*np.pi,100))

# def CircInhomMaq(model,x0=0,y0=0,R=1,order=1,kaq=[1],z=[1,0],c=[],Saq=[0.001],Sll=[0],topboundary='imp',phreatictop=False,label=None,test=False):
#     CircInhomDataMaq(model,x0,y0,R,kaq,z,c,Saq,Sll,topboundary,phreatictop)
#     return CircInhom(model,x0,y0,R,order,label,test)

# def CircInhom3D(model,x0=0,y0=0,R=1,order=1,kaq=[1,1,1],z=[4,3,2,1],Saq=[0.3,0.001,0.001],kzoverkh=[.1,.1,.1],phreatictop=True,label=None):
#     CircInhomData3D(model,x0,y0,R,kaq,z,Saq,kzoverkh,phreatictop)
#     return CircInhom(model,x0,y0,R,order,label)


#

# ml = ttim.ModelMaq(kaq=[4,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=1,tmax=10,M=20)
##ls = MscreenLineSinkDitchString(ml,[(-1,0),(0,0),(1,0)],tsandQ=[(0.0,1.0)],layers=[2])
# e1a = EllipseInhomDataMaq(ml,0,0,along=2.0,bshort=1.0,angle=0.0,kaq=[10,2],z=[4,2,1,0],c=[200],Saq=[2e-3,2e-4],Sll=[1e-5])
# e1 = EllipseInhom(ml,0,0,along=2.0,bshort=1.0,angle=0.0,order=5)
# e1 = EllipseInhomMaq(ml,0,0,along=2.0,bshort=1.0,angle=0.0,order=5,kaq=[10,2],z=[4,2,1,0],c=[200],Saq=[2e-3,2e-4],Sll=[1e-5])
## Same inside and outside
# c1 = CircInhomMaq(ml,0,0,2.0,order=5,kaq=[4,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6])
# c1 = CircInhomMaq(ml,0,0,2.0,order=5,kaq=[10,.1],z=[4,2,1,0],c=[200],Saq=[2e-3,2e-4],Sll=[1e-5])
##c2 = CircInhomMaq(ml,0,0,5000.0,order=1,kaq=[10,2],z=[4,2,1,0],c=[200],Saq=[2e-3,2e-4],Sll=[1e-5])
##ml.initialize()
##c2.circ_in_small[:] = 0
##c2.circ_out_small[:] = 0
# w = DischargeWell(ml,xw=.5,yw=0,rw=.1,tsandQ=[0,5.0],layers=1)
# ml.solve()

# ml.solve()
# h1,h2 = np.zeros((2,e1.Ncp)), np.zeros((2,e1.Ncp))
# qn1,qn2 = np.zeros((2,e1.Ncp)), np.zeros((2,e1.Ncp))
# for i in range(e1.Ncp):
#    h1[:,i] = ml.head(e1.xc[i],e1.yc[i],2,aq=e1.aqin)[:,0]
#    h2[:,i] = ml.head(e1.xc[i],e1.yc[i],2,aq=e1.aqout)[:,0]
#    qx1,qy1 = ml.discharge(e1.xc[i],e1.yc[i],2,aq=e1.aqin)
#    qx2,qy2 = ml.discharge(e1.xc[i],e1.yc[i],2,aq=e1.aqout)
#    a = e1a.outwardnormalangle(e1.xc[i],e1.yc[i])
#    qn1[:,i] = qx1[:,0]*np.cos(a) + qy1[:,0]*np.sin(a)
#    qn2[:,i] = qx2[:,0]*np.cos(a) + qy2[:,0]*np.sin(a)


# ml = ttim.ModelMaq(kaq=[10,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=.1,tmax=10)
# w1 = ttim.Well(ml,0,2,.1,tsandQ=[(0,10)],layers=[1])
# ls2 = ZeroHeadLineSinkString(ml,xy=[(-10,-2),(0,-4),(4,0)],layers=[1])
# ls1 = MscreenLineSinkDitchString(ml,xy=[(-10,0),(0,0),(10,10)],tsandQ=[(0.0,7.0)],res=0.0,wh='H',layers=[2],label=None)
# ml.solve()

# ml = ttim.ModelMaq([1,20,2],[25,20,18,10,8,0],c=[1000,2000],Saq=[0.1,1e-4,1e-4],Sll=[0,0],phreatictop=True,tmin=1e-6,tmax=10,M=30)
# w1 = ttim.Well(ml,0,0,.1,tsandQ=[(0,1000)],layers=[2])
# ls1 = ZeroMscreenLineSink(ml,10,-5,10,5,layers=[1,2,3],res=0.5,wh=1,vres=3,wv=1)
# w2 = ZeroMscreenWell(ml,10,0,res=1.0,layers=[1,2,3],vres=1.0)
# w3 = ttim.Well(ml,0,-10,.1,tsandQ=[(0,700)],layers=[2])
# ml.solve()
##ml1 = ttim.ModelMaq([1,20,2],[25,20,18,10,8,0],c=[1000,2000],Saq=[1e-4,1e-4,1e-4],Sll=[0,0],tmin=0.1,tmax=10000,M=30)
##w1 = ttim.Well(ml1,0,0,.1,tsandQ=[(0,1000)],layers=[2],res=0.1)
##ml1.solve()
# t = np.logspace(-1,3,100)
# h0 = ml.head(50,0,t)
##h1 = ml1.head(50,0,t)
##w = MscreenWell(ml,0,0,.1,tsandQ=[(0,1000),(100,0),(365,1000),(465,0)],layers=[2,3])
##w2 = HeadWell(ml,50,0,.2,tsandh=[(0,1)],layers=[2])
##y = [-500,-300,-200,-100,-50,0,50,100,200,300,500]
##x = 50 * np.ones(len(y))
##ls = ZeroHeadLineSinkString(ml,xy=zip(x,y),layers=[1])
##w = ttim.Well(ml,0,0,.1,tsandQ=[(0,1000),(100,0)],layers=[2])
##ml.solve()


# ml = ttim.Model3D( kaq=[2,1,5,10,4], z=[10,8,6,4,2,0], Saq=[.1,.0001,.0002,.0002,.0001], phreatictop=True, kzoverkh=0.1, tmin=1e-3, tmax=1e3 )
# w = MscreenWell(ml,0,-25,rw=.3,tsandQ=[(0,100),(100,50)],layers=[2,3])
# ml.solve()

##ml = ttim.Model3D(kaq=2.0,z=[10,5,0],Saq=[.002,.001],kzoverkh=0.2,phreatictop=False,tmin=.1,tmax=10,M=15)
# ml = ttim.ModelMaq(kaq=[10,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=100,tmax=300,M=50)
# w = HeadWellNew(ml,0,0,.1,tsandh=[(0.0,1.0)],layers=1)
# ml.solve()
##L1 = np.sqrt(10**2+5**2)
##ls1 = LineSink(ml,-10,-10,0,-5,tsandQ=[(0,.05*L1),(1,.02*L1)],res=1.0,layers=[1,2],label='mark1')
# w = MscreenWell(ml,-5,-5,.1,[0,5],layers=[1,2])
# L2 = np.sqrt(10**2+15**2)
# ls2 = LineSink(ml,0,-5,10,10,tsandQ=[(0,.03*L2),(2,.07*L2)],layers=[1],label='mark2')
##ls3a = ZeroHeadLineSink(ml,-10,5,-5,5,res=1.0,layers=[1,2])
##ls3b = ZeroHeadLineSink(ml,-5,5,0,5,res=1.0,layers=[1,2])
##ls3c = ZeroHeadLineSink(ml,0,5,5,5,res=1.0,layers=[1,2])
##lss = HeadLineSinkString(ml,[(-10,5),(-5,5),(0,5)],tsandh=[(0,0.02),(3,0.01)],res=1.0,layers=[1,2])
# lss = ZeroHeadLineSinkString(ml,[(-10,5),(-5,5),(0,5),(5,5)],res=1.0,layers=[1,2])
##lss = MscreenLineSinkString(ml,[(-10,5),(-5,5),(0,5)],tsandQ=[(0,0.2),(3,0.1)],res=1.0,layers=[1,2])
##lss = ZeroMscreenLineSinkString(ml,[(-10,5),(-5,5),(0,5)],res=1.0,layers=[1,2])
##ml.initialize()
# ml.solve()
# print ml.potential(50,50,[0.5,5])

# ml2 = ttim.ModelMaq(kaq=[10,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=.1,tmax=10,M=15)
# L1 = np.sqrt(10**2+5**2)
# ls1b = LineSink(ml2,-10,-10,0,-5,tsandQ=[(0,.05*L1),(1,.02*L1)],res=1.0,layers=[1,2],label='mark1')
# L2 = np.sqrt(10**2+15**2)
# ls2b = LineSink(ml2,0,-5,10,10,tsandQ=[(0,.03*L2),(2,.07*L2)],layers=[1],label='mark2')
##ls3a = HeadLineSink(ml2,-10,5,-5,5,tsandh=[(0,0.02),(3,0.01)],res=1.0,layers=[1,2])
##ls3b = HeadLineSink(ml2,-5,5,0,5,tsandh=[(0,0.02),(3,0.01)],res=1.0,layers=[1,2])
##ls3a = ZeroHeadLineSink(ml2,-10,5,-5,5,res=1.0,layers=[1,2])
##ls3b = ZeroHeadLineSink(ml2,-5,5,0,5,res=1.0,layers=[1,2])
##ls3a = MscreenLineSink(ml2,-10,5,-5,5,tsandQ=[(0,0.2),(3,0.1)],res=1.0,layers=[1,2])
##ls3b = MscreenLineSink(ml2,-5,5,0,5,tsandQ=[(0,0.2),(3,0.1)],res=1.0,layers=[1,2])
# ls3a = ZeroMscreenLineSink(ml2,-10,5,-5,5,res=1.0,layers=[1,2])
# ls3b = ZeroMscreenLineSink(ml2,-5,5,0,5,res=1.0,layers=[1,2])
##lssb = HeadLineSinkStringOld(ml2,[(-10,5),(-5,5),(0,5)],tsandh=[(0,0.02),(3,0.01)],res=0.0,layers=[1,2])
# ml2.solve()
# print ml2.potential(50,50,[0.5,5])

# lss = HeadLineSinkString(ml,[(-10,5),(-5,5),(0,5)],tsandh=[(0,0.02),(3,0.01)],res=1.0,layers=[1,2])
# lss = MscreenLineSinkString(ml,[(-10,5),(-5,5),(0,5)],tsandQ=[(0,.03*5),(2,.07*5)],res=0.5,layers=[1,2])
# ls3a = MscreenLineSink(ml,-10,5,-5,5,tsandQ=[(0,.03*5),(2,.07*5)],res=0.5,layers=[1,2])
# ls3b = MscreenLineSink(ml,-5,5,0,5,tsandQ=[(0,.03*5),(2,.07*5)],res=0.5,layers=[1,2])
#
# ml2 = ttim.ModelMaq(kaq=[10,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=.1,tmax=10,M=15)
# L1 = np.sqrt(10**2+5**2)
# ls1a = LineSink(ml2,-10,-10,0,-5,tsandQ=[(0,.05*L1),(1,.02*L1)],res=1.0,layers=[1,2],label='mark1')
# L2 = np.sqrt(10**2+15**2)
# ls2a = LineSink(ml2,0,-5,10,10,tsandQ=[(0,.03*L2),(2,.07*L2)],layers=[1],label='mark2')
# ls3a = HeadLineSink(ml2,-10,5,-5,5,tsandh=[(0,0.02),(3,0.01)],res=1.0,layers=[1,2])
# ls3b = HeadLineSink(ml2,-5,5,0,5,tsandh=[(0,0.02),(3,0.01)],res=1.0,layers=[1,2])


##lss = HeadLineSinkString(ml,[(-10,5),(-5,5),(0,5)],tsandh=[(0,0.02),(3,0.01)],res=0.0,layers=[1,2])
##ls3 = ZeroMscreenLineSink(ml,-10,5,0,5,res=1.0,layers=[1,2])
# ml = ttim.ModelMaq(kaq=[10,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=.1,tmax=10,M=15)
# w1 = ttim.Well(ml,0,0,.1,tsandQ=[(0,5),(1,2)],res=1.0,layers=[1,2])
# w2 = ttim.Well(ml,100,0,.1,tsandQ=[(0,3),(2,7)],layers=[1])
##w3 = MscreenWell(ml,0,100,.1,tsandQ=[(0,2),(3,1)],res=2.0,layers=[1,2])
# w3 = ZeroMscreenWell(ml,0,100,.1,res=2.0,layers=[1,2])
##w3 = ZeroHeadWell(ml,0,100,.1,res=1.0,layers=[1,2])
##w3 = HeadWell(ml,0,100,.1,tsandh=[(0,2),(3,1)],res=1.0,layers=[1,2])
# ml.solve()
###print ml.potential(2,3,[.5,5])
# print ml.potential(50,50,[0.5,5])
# ml2.solve()
# print ml2.potential(50,50,[.5,5])
# print lss.strength([.5,5])
#
# ml2 = ttim.ModelMaq(kaq=[10,5],z=[4,2,1,0],c=[100],Saq=[1e-3,1e-4],Sll=[1e-6],tmin=0.1,tmax=10,M=15)
# ls1a = LineSink(ml2,-10,-10,0,-5,tsandsig=[(0,.05),(1,.02)],res=1.0,layers=[1,2],label='mark1')
# ls2a = LineSink(ml2,0,-5,10,10,tsandsig=[(0,.03),(2,.07)],layers=[1],label='mark2')
# ls3a = HeadLineSinkStringOld(ml2,[(-10,5),(-5,5),(0,5)],tsandh=[(0,0.02),(3,0.01)],res=0.0,layers=[1,2])
# ml2.solve()
# print ml2.potential(50,50,[0.5,5])

# print 'Q from strength:  ',w3.strength(.5)
# print 'Q from head diff: ',(ml.head(w3.xc,w3.yc,.5)-w3.headinside(.5))/w3.res*2*np.pi*w3.rw*ml.aq.Haq[:,np.newaxis]
# print 'Q from head diff: ',(ml.head(w3.xc,w3.yc,.5)-2.0)/w3.res*2*np.pi*w3.rw*ml.aq.Haq[:,np.newaxis]
# print w3.strength([.5,5])
# print ls3.strength([.5,5])
# print sum(ls3.strength([.5,5]),0)
# Q = w3.strength([.5,5])
# print sum(Q,0)
# print ml.potential(w3.xc,w3.yc,[.5,5])
